/**
 * Copyright (C) 2012 Michael Oestergaard Pedersen.
 * Distributed under the GNU GPL v3. For full terms see the file LICENSE.txt.
 *
 * This file is part of VerifiedQR.
 *
 * VerifiedQR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VerifiedQR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VerifiedQR.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package dk.bytopia.crypto.signature.bls;

import org.bouncycastle.crypto.digests.GeneralDigest;
import org.bouncycastle.crypto.digests.RIPEMD160Digest;
import org.bouncycastle.crypto.digests.SHA224Digest;
import org.apache.commons.codec.binary.Base64;
import it.unisa.dia.gas.jpbc.Element;
import it.unisa.dia.gas.jpbc.Pairing;
import it.unisa.dia.gas.plaf.jpbc.pairing.PairingFactory;
import it.unisa.dia.gas.plaf.jpbc.field.curve.CurveElement;
import java.io.Writer;
import java.io.OutputStreamWriter;
import java.io.FileOutputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.UnsupportedEncodingException;
import java.io.IOException;
import java.util.Scanner;
import java.util.StringTokenizer;
import java.security.InvalidKeyException;
import java.security.SignatureException;
import java.security.InvalidParameterException;

public class BLS {

    private Pairing pairing = null;
    private Element pk = null;
    private Element sk = null;
    private Element g = null;
    private GeneralDigest digest = null;
    private int signatureSize = 0;

    public BLS(String curveProperties) throws InvalidParameterException {
	pairing = PairingFactory.getPairing(curveProperties);
	int fieldSize = pairing.getG1().getLengthInBytes() / 2;

	switch (fieldSize * 8) {
	case 160:  digest = new RIPEMD160Digest();
	    break;
	case 224:  digest = new SHA224Digest();
	    break;
	default:
	    throw new InvalidParameterException("G1 must be a curve over a field of size 160 or 224 bits.");
	}

	signatureSize = fieldSize + 1;
    }

    public void keygen() {
        g = pairing.getG2().newRandomElement();
        sk = pairing.getZr().newRandomElement();
        pk = g.duplicate().powZn(sk);
    }

    public void loadPublicKey(String filename) throws InvalidKeyException {
	try {
	    g = null;
	    pk = null;
	    Scanner scanner = new Scanner(new FileInputStream(filename), "ISO-8859-1");
	    try {
		while (scanner.hasNextLine()) {
		    StringTokenizer st = new StringTokenizer(scanner.nextLine(), ":");
		    if (st.hasMoreTokens()) {
			String tok = st.nextToken();
			if (st.hasMoreTokens()) {
			    if (tok.equals("g")) {
				g = pairing.getG2().newElement();
				g.setFromBytes(Base64.decodeBase64(st.nextToken().getBytes("ISO-8859-1")));
			    }
			    else if (tok.equals("pk")) {
				pk = pairing.getG2().newElement();
				pk.setFromBytes(Base64.decodeBase64(st.nextToken().getBytes("ISO-8859-1")));
			    }
			}
		    }
		    if (g != null && pk != null) break;
		}
		if (g == null || pk == null)
		    throw new InvalidKeyException("No public key found in '" + filename + "'.");
	    }
	    finally{
		scanner.close();
	    }
	} catch (FileNotFoundException e) {
	    throw new InvalidKeyException(e.getMessage());
	} catch (UnsupportedEncodingException e) {
	    throw new InvalidKeyException(e.getMessage());
	}
    }

    public void storePublicKey(String filename) throws InvalidKeyException {
	if (g == null || pk == null)
	    throw new InvalidKeyException("Public key not initialised.");

	try {
	    Writer out = new OutputStreamWriter(new FileOutputStream(filename), "ISO-8859-1");
	    try {
		out.write("g:" + new String(Base64.encodeBase64(g.toBytes()), "ISO-8859-1") + "\n");
		out.write("pk:" + new String(Base64.encodeBase64(pk.toBytes()), "ISO-8859-1") + "\n");
	    }
	    finally {
		out.close();
	    }
	} catch (FileNotFoundException e) {
	    throw new InvalidKeyException(e.getMessage());
	} catch (UnsupportedEncodingException e) {
	    throw new InvalidKeyException(e.getMessage());
	} catch (IOException e) {
	    throw new InvalidKeyException(e.getMessage());
	}	
    }

    public void loadSecretKey(String filename) throws InvalidKeyException {
	try {
	    sk = null;
	    Scanner scanner = new Scanner(new FileInputStream(filename), "ISO-8859-1");
	    try {
		while (scanner.hasNextLine()) {
		    StringTokenizer st = new StringTokenizer(scanner.nextLine(), ":");
		    if (st.hasMoreTokens()) {
			String tok = st.nextToken();
			if (st.hasMoreTokens() && tok.equals("sk")) {
			    sk = pairing.getZr().newElement();
			    sk.setFromBytes(Base64.decodeBase64(st.nextToken().getBytes("ISO-8859-1")));
			}
		    }
		    if (sk != null) break;
		}
		if (sk == null)
		    throw new InvalidKeyException("No secret key found in '" + filename + "'.");
	    }
	    finally{
		scanner.close();
	    }
	} catch (FileNotFoundException e) {
	    throw new InvalidKeyException(e.getMessage());
	} catch (UnsupportedEncodingException e) {
	    throw new InvalidKeyException(e.getMessage());
	}
    }

    public void storeSecretKey(String filename) throws InvalidKeyException {
	if (sk == null)
	    throw new InvalidKeyException("Secret key not initialised.");

	try {
	    Writer out = new OutputStreamWriter(new FileOutputStream(filename), "ISO-8859-1");
	    try {
		out.write("sk:" + new String(Base64.encodeBase64(sk.toBytes()), "ISO-8859-1") + "\n");
	    }
	    finally {
		out.close();
	    }
	} catch (FileNotFoundException e) {
	    throw new InvalidKeyException(e.getMessage());
	} catch (UnsupportedEncodingException e) {
	    throw new InvalidKeyException(e.getMessage());
	} catch (IOException e) {
	    throw new InvalidKeyException(e.getMessage());
	}
    }

    public int getSignatureSize() {
	return signatureSize;
    }

    public byte[] sign(byte[] message) throws InvalidKeyException {
	if (sk == null)
	    throw new InvalidKeyException("Secret key not initialised.");
	else if (g == null || pk == null)
	    throw new InvalidKeyException("Public key not initialised.");

	// Get a hash of the message
	byte[] hash = hashMessage(message);
	
	// Create an element in G1 from the hash
        Element h = pairing.getG1().newElement().setFromHash(hash, 0, hash.length);

        // Generate the signature
        Element sig = h.powZn(sk);

	return ((CurveElement)sig).toBytesCompressed();
    }

    public void verify(byte[] message, byte[] sig) throws InvalidKeyException, SignatureException {
	if (g == null || pk == null)
	    throw new InvalidKeyException("Public key not initialised.");

	if (sig.length != signatureSize)
	    throw new SignatureException("Invalid signature size.");

	// Get a hash of the message
	byte[] hash = hashMessage(message);

	// Create an element in G1 from the hash
        Element h = pairing.getG1().newElement().setFromHash(hash, 0, hash.length);

	// Verify the signature
	Element temp1 = pairing.getG1().newElement();
	((CurveElement)temp1).setFromBytesCompressed(sig);
        Element temp2 = pairing.pairing(temp1, g);
        Element temp3 = pairing.pairing(h, pk);

        if (!temp2.isEqual(temp3))
	    throw new SignatureException("Invalid signature.");
    }

    private byte[] hashMessage(byte[] message) {
	digest.update(message, 0, message.length);
	byte[] hash = new byte[digest.getDigestSize()];
	digest.doFinal(hash, 0);
	return hash;
    }
}
