/**
 * Copyright (C) 2012 Michael Oestergaard Pedersen.
 * Distributed under the GNU GPL v3. For full terms see the file LICENSE.txt.
 *
 * This file is part of VerifiedQR.
 *
 * VerifiedQR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VerifiedQR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VerifiedQR.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package dk.bytopia.verifiedqr;

import java.util.Vector;
import java.io.UnsupportedEncodingException;
import java.util.zip.Inflater;
import java.util.zip.Deflater;
import java.util.zip.DataFormatException;
import java.util.Arrays;

class DataRecord {

    private final int MAX_ELEMENTS = 127;
    private final int MAX_LENGTH = 255;
    private final int COMPRESS_BUFFER_SIZE = 65536;

    private int totalLength = 0;

    private Vector<String> values = new Vector<String>();

    private byte[] compress(byte[] input) throws DataFormatException {
	byte[] output = new byte[COMPRESS_BUFFER_SIZE];
	Deflater compresser = new Deflater(9, true);

	compresser.setInput(input, 1, input.length - 1);
	compresser.finish();

	int compressedDataLength = compresser.deflate(output, 1, output.length - 1);
	if (compressedDataLength == output.length - 1) {
	    throw new DataFormatException("Compressed data too large for output buffer.");
	}

	if (compressedDataLength + 1 < input.length) {
	    output[0] = (byte)(input[0] + 0x80);
	    return Arrays.copyOf(output, compressedDataLength + 1);
	} else {
	    return input;
	}
    }

    private byte[] decompress(byte[] input) throws DataFormatException {
	if ((input[0] & 0x80) == 0) {
	    return input;
	}

	byte[] output = new byte[COMPRESS_BUFFER_SIZE];
	Inflater decompresser = new Inflater(true);

	output[0] = input[0];
	decompresser.setInput(input, 1, input.length - 1);
	int resultLength = decompresser.inflate(output, 1, output.length - 1);
	decompresser.end();

	if (resultLength == output.length - 1) {
	    throw new DataFormatException("Decompressed data too large for output buffer.");
	}

	return Arrays.copyOf(output, resultLength + 1);
    }

    public void setFromVector(Vector<String> data) throws IllegalArgumentException {
	if (data != null) {
	    if (data.size() > MAX_ELEMENTS)
		throw new IllegalArgumentException("Too many values: The vector can only contain at most " + MAX_ELEMENTS + " elements.");
	    
	    byte[] temp;
	    totalLength = 0;
	    try {
		for (int i = 0; i < data.size(); i++) {
		    temp = data.elementAt(i).getBytes("UTF8");
		    totalLength = totalLength + temp.length;
		    if (temp.length <= MAX_LENGTH) {
			values.add(data.elementAt(i));
		    } else {
			throw new IllegalArgumentException("Invalid value: An element can only contain " + MAX_LENGTH + " bytes.");
		    }
		}
	    } catch (UnsupportedEncodingException e) {
		throw new IllegalArgumentException(e.getMessage());
	    }
	}
    }

    public void setFromBytes(byte[] data) throws IllegalArgumentException {
	int size = data[0] & 0x7F;
	values = new Vector<String>(size);

	try {
	    data = decompress(data);
	} catch (DataFormatException e) {
	    throw new IllegalArgumentException(e.getMessage());
	}

	int strLength = 0;
	int offset = size + 1;
	totalLength = 0;
	try {
	    for (int i = 0; i < size; i++) {
		strLength = data[i+1] & 0xFF;
		totalLength = totalLength + strLength;
		if (offset + strLength > data.length)
		    throw new IllegalArgumentException("Invalid data structure: Attempt to read past the end of the buffer.");
		values.add(new String(data, offset, strLength, "UTF8"));
		offset = offset + strLength;
	    }
	} catch (UnsupportedEncodingException e) {
	    throw new IllegalArgumentException(e.getMessage());
	}
    }

    public Vector<String> getVector() {
	return new Vector<String>(values);
    }

    public byte[] getBytes() {
	int valCount = values.size();

	int bufSize = 1 + valCount + totalLength;
	byte[] data = new byte[bufSize];
	data[0] = (byte)(valCount & 0x7F);

	byte[] val;
	int valLength = 0;
	int offset = 1 + valCount;
	try {
	    for (int i = 0; i < valCount; i++) {
		val = values.elementAt(i).getBytes("UTF8");
		valLength = val.length;
		data[i+1] = (byte)(valLength & 0xFF);
		System.arraycopy(val, 0, data, offset, valLength);
		offset = offset + valLength;
	    }
	} catch (UnsupportedEncodingException e) {
	    System.err.println("Failed to convert string to UTF-8");
	    return null;
	}

	try {
	    data = compress(data);
	} catch (DataFormatException e) {
	    throw new IllegalArgumentException(e.getMessage());
	}

	return data;
    }

    public String toString() {
	return values.toString();
    }
}
