/**
 * Copyright (C) 2012 Michael Oestergaard Pedersen.
 * Distributed under the GNU GPL v3. For full terms see the file LICENSE.txt.
 *
 * This file is part of VerifiedQR.
 *
 * VerifiedQR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VerifiedQR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VerifiedQR.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package dk.bytopia.verifiedqr;

import dk.bytopia.crypto.signature.bls.BLS;
import org.apache.commons.codec.binary.Base64;
import java.util.Vector;
import java.io.File;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.security.InvalidKeyException;
import java.security.SignatureException;
import com.google.zxing.BarcodeFormat;
import com.google.zxing.Writer;
import com.google.zxing.WriterException;
import com.google.zxing.common.BitMatrix;
import com.google.zxing.qrcode.QRCodeWriter;
import com.google.zxing.client.j2se.MatrixToImageWriter;

public class VerifiedQR {

    public String encodeAndSign(BLS bls, Vector<String> values) throws IllegalArgumentException, InvalidKeyException {
	DataRecord data = new DataRecord();
	data.setFromVector(values);

	byte[] msg = data.getBytes();
	byte[] sig = bls.sign(msg);
	byte[] out = new byte[sig.length + msg.length];

	System.arraycopy(sig, 0, out, 0, sig.length);
	System.arraycopy(msg, 0, out, sig.length, msg.length);

	try {
	    return new String(Base64.encodeBase64(out), "ISO-8859-1");
	} catch (UnsupportedEncodingException e) {
	    throw new IllegalArgumentException("Invalid Base64 encoded data.");
	}
    }

    public Vector<String> decodeAndVerify(BLS bls, String data) throws IllegalArgumentException, InvalidKeyException, SignatureException {
	byte[] buf = null;
	try {
	    buf = Base64.decodeBase64(data.getBytes("ISO-8859-1"));
	} catch (UnsupportedEncodingException e) {
	    throw new IllegalArgumentException("Invalid Base64 encoded data.");
	}

	int sigLength = bls.getSignatureSize();
	if (buf == null || sigLength > buf.length)
	    throw new IllegalArgumentException("Buffer too small to be valid.");

	int msgLength = buf.length - sigLength;
	byte[] sig = new byte[sigLength];
	byte[] msg = new byte[msgLength];

	System.arraycopy(buf, 0, sig, 0, sigLength);
	System.arraycopy(buf, sigLength, msg, 0, msgLength);

	bls.verify(msg, sig);

	DataRecord out = new DataRecord();
	out.setFromBytes(msg);

	return out.getVector();
    }

    public void generate(String filename, String data, int width, int height) throws IOException {
	BitMatrix matrix = null;
	Writer writer = new QRCodeWriter();
	try {
	    matrix = writer.encode(data, BarcodeFormat.QR_CODE, width, height);
	} catch (WriterException e) {
	    System.out.println(e.getMessage());
	}

	File file = new File(filename);
	MatrixToImageWriter.writeToFile(matrix, "PNG", file);
    }
}
